/**
 * libFAD - Flash Animation Decode library
 * Copyright (C) 2005-2006 VGSystem Technologies, Inc.
 *
 * libFAD is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: fad.render.c,v 1.32 2006/03/07 10:17:57 wrxzzj Exp $
 */
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include "fad.render.h"

void default_render_init(fad_render_t *render) {
  /**do nothing*/
}

void default_render_term(fad_render_t *render) {
  /**do nothing*/
}

s32_t default_render_open_snddev(fad_render_t *render, u8_t sndattr) {
  u32_t rates[] = {5512, 11025, 22050, 44100}, rate;
  u8_t nchannels = 1;
  s16_t format = AFMT_U8;

  if(sndattr == render->sndattr)
    return 0;

  render->sndattr = sndattr;
  if(sndattr&0x02)
    format = AFMT_S16_LE;
  if(sndattr&0x01)
    nchannels = 2;

  if(render->snddev)
    close(render->snddev);
  render->snddev = 0;

  rate = rates[(sndattr&0x0c)>>2];
  render->snddev = open("/dev/dsp", O_WRONLY, 0);
  if(render->snddev == -1) {
    FAD_ERROR("open sound device(/dev/dsp) failure.\n");
    goto snddev_error;
  }

  if(ioctl(render->snddev, SNDCTL_DSP_SETFMT, &format) < 0) {
    FAD_ERROR("set sound device format failure.\n");
    goto snddev_error;
  }

  if(ioctl(render->snddev, SNDCTL_DSP_CHANNELS, &nchannels) == -1) {
    FAD_ERROR("set sound device nchannels failure.\n");
    goto snddev_error;
  }

  if(ioctl(render->snddev, SNDCTL_DSP_SPEED, &rate) == -1) {
    FAD_ERROR("set sound device rate failure.\n");
    goto snddev_error;
  }

  FAD_ERROR("open /dev/dsp,  nchannels = %d, rate = %d, attr = %x\n", nchannels, rate, sndattr);

  return 0;

snddev_error:
  if(render->snddev)
    close(render->snddev);
  render->snddev = 0;

  return -1;
}

s32_t default_render_write_pcm(fad_render_t *render, void *buffer, s32_t size) {
  if(render->snddev)
    return write(render->snddev, buffer, size);

  return -1;
}

void default_render_close_snddev(fad_render_t *render) {
  if(render->snddev) {
    close(render->snddev);
    render->snddev = 0;
  }
}
